# -*-: coding utf-8 -*-
""" Helper methods for OS related tasks. """

import os
import shlex
import subprocess
import urllib2


def cmd_exists(cmd):
    """ Check if a command exists.

    :param cmd: the command to look for.
    :return: true if the command exists, false otherwise.
    """
    return subprocess.call("type " + cmd, shell=True,
                           stdout=subprocess.PIPE, stderr=subprocess.PIPE) == 0


def is_raspi_os():
    """ Check if the current system is Raspberry.

    :return: true if the current system is Raspberry.
    """
    return 'arm' in " ".join(os.uname())


def execute_command(command):
    """ Execute a shell command.

    :param command: the command to execute.
    """
    subprocess.Popen(command.split(), stdout=subprocess.PIPE).communicate()


def pipe_commands(first_command, second_command, silent):
    """ Execute piped commands: `first_command | second_command`.

    :param first_command: the first command to execute.
    :param second_command: the second command to execute.
    """
    process1 = subprocess.Popen(first_command.split(), stdout=subprocess.PIPE)
    if silent:
        FNULL = open(os.devnull, 'w')
        process2 = subprocess.Popen(
            second_command.split(), stdin=process1.stdout, stdout=FNULL)
    else:
        process2 = subprocess.Popen(
            second_command.split(), stdin=process1.stdout)
    process1.stdout.close()
    process2.communicate()


def remove_file(file_path):
    """ Delete a file.

    :param file_path: the path to the file.
    """
    try:
        os.remove(file_path)
    except OSError:
        pass


def download_file(url, output_file):
    """ Download a file.

    :param url: the remote location of the file.
    :param output_file: the file to write to.
    """
    downloaded_file = urllib2.urlopen(url)
    with open(output_file, 'wb') as output:
        output.write(downloaded_file.read())


def ask_yes_no(question):
    """ Ask a yes/no question in the prompt.

    :param question: the question to ask.
    :return: true if the user answered yes (or empty), false otherwise
    """
    answer = raw_input("{} [Y/n] ".format(question))
    if answer is not None and answer.strip() != "" and answer.lower() != "y":
        return False
    return True


def which(command):
    """ Get full path for an executable.

    :param command: the executable command, e.g. 'node'.
    :return: the full path for the command, e.g. '/usr/local/bin/node'.
    """
    try:
        return subprocess.check_output(
            ['which', command]).strip()
    except subprocess.CalledProcessError:
        return None



# -*-: coding utf-8 -*-
""" Intent definition from a YAML config. """

# pylint: disable=too-few-public-methods
class IntentDef:
    """ Intent definition from a YAML config. """

    def __init__(self, name, action):
        """ Initialisation.

        :param name: the name of the intent.
        :param action: the code to execute.
        """
        self.name = name
        self.action = action



# -*-: coding utf-8 -*-
"""The install command."""

import os
import shutil

from .base import Base, ASSISTANT_DIR, ASSISTANT_ZIP_FILENAME, \
    ASSISTANT_ZIP_PATH, INTENTS_DIR, SNIPSFILE

from ..utils.snipsfile_parser import Snipsfile, SnipsfileParseException, \
    SnipsfileNotFoundError

from ..utils.assistant_downloader import AssistantDownloader, \
    AssistantDownloaderException
from ..utils.intent_class_generator import IntentClassGenerator
from ..utils.pip_installer import PipInstaller
from ..utils.snips import Snips, SnipsUnsupportedPlatform, SnipsInstallationFailure
from ..utils.os_helpers import cmd_exists, is_raspi_os, remove_file
from ..utils.microphone_setup import MicrophoneSetup
from ..utils.systemd import Systemd
from ..utils.bluetooth import Bluetooth

from snipsskillscore.logging import log, log_success, log_error


# pylint: disable=too-few-public-methods
class Install(Base):
    """The install command."""

    def run(self):
        """ Command runner. """
        try:
            snipsfile = Snipsfile(SNIPSFILE)
        except SnipsfileNotFoundError:
            log_error("Snipsfile not found. Please create one.")
            return
        except SnipsfileParseException as err:
            log_error(err)
            return

        if self.options['bluetooth'] == True:
            Install.setup_bluetooth(snipsfile.mqtt_hostname, snipsfile.mqtt_port)
            return

        if not Snips.is_installed():
            try:
                Snips.install()
            except SnipsUnsupportedPlatform:
                log_error("Currently, the Snips SDK only runs on a Raspberry Pi. " +
                          "Skipping installation of the Snips SDK. " +
                          "If you wish to install the Snips SDK, " +
                          "run this command from a Raspberry Pi.")
            except SnipsInstallationFailure as e:
                log_error("Error installing Snips {}".format(e))

        if snipsfile.assistant_url is None:
            log_error("No assistants found in Snipsfile.")
            return

        log("Fetching assistant.")
        try:
            AssistantDownloader.download(snipsfile.assistant_url,
                                         ASSISTANT_DIR,
                                         ASSISTANT_ZIP_FILENAME)
        except AssistantDownloaderException:
            log_error("Error downloading assistant. " +
                      "Make sure the provided URL in the Snipsfile is correct, " +
                      "and that there is a working network connection.")
            return

        if Snips.is_installed():
            log("Loading Snips assistant.")
            Snips.load_assistant(ASSISTANT_ZIP_PATH)

        log("Generating definitions.")
        try:
            shutil.rmtree(INTENTS_DIR)
        except Exception:
            pass

        if is_raspi_os():
            log("Setting up microphone.")
            MicrophoneSetup.setup(snipsfile.microphone_config)
        else:
            log("System is not Raspberry Pi. Skipping microphone setup.")

        generator = IntentClassGenerator()
        generator.generate(ASSISTANT_ZIP_PATH, INTENTS_DIR)

        if snipsfile.skilldefs is not None and len(snipsfile.skilldefs) > 0:
            log("Installing skills.")
            for skill in snipsfile.skilldefs:
                log("Installing {}.".format(skill.package_name))
                PipInstaller.install(skill.package_name)

        if is_raspi_os():
            Systemd.setup()

        Install.setup_bluetooth(snipsfile.mqtt_hostname, snipsfile.mqtt_port)

        remove_file(ASSISTANT_ZIP_PATH)

        log_success("All done! Type 'snipsskills run' to launch the skills server. " +
                    "Make sure you have a running instance of the Snips SDK. " +
                    "If you have set up Snips Skills as a systemd service, " +
                    "you can also reboot your device " +
                    "and it will be run automatically at launch.")

    @staticmethod
    def setup_bluetooth(mqtt_hostname, mqtt_port):
        if not is_raspi_os():
            log("System is not Raspberry Pi. Skipping Bluetooth setup.")
            return
        Bluetooth.setup(mqtt_hostname, mqtt_port)



# -*-: coding utf-8 -*-
""" Tools to automatically generate intent classes from an assistant
    definition.
"""

import os
import re
import json
import zipfile

from jinja2 import Environment, PackageLoader


def camel_case_to_underscore(text):
    """ Convert camel-case to underscore.

    :param text: a text, potentially in camel-case format.
    :return: the text, converted to underscore format.
    """
    underscored = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', text)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', underscored).lower()


def to_camelcase_capitalized(text):
    """ Convert to camel-case and capitalize.

    :param text: a text, potentially with dashes and underscores.
    :return: the text, converted to cancel-case format, and capitalized.
    """
    hyphens = re.sub(r'(?!^)-([a-zA-Z])',
                     lambda m: m.group(1).upper(), text)
    underscores = re.sub(r'(?!^)_([a-zA-Z])',
                         lambda m: m.group(1).upper(), hyphens)
    return underscores[:1].upper() + underscores[1:]


def save_intent_file(output_dir, filename, text):
    """ Save a text string to a given file.

    :param filename: a file name.
    :param text: a text to save.
    """
    if not os.path.exists(output_dir):
        os.makedirs(output_dir)
    if not os.path.exists("{}/intents".format(output_dir)):
        os.makedirs("{}/intents".format(output_dir))
    output_filename = "{}/intents/{}".format(output_dir, filename)
    with open(output_filename, "w") as output_file:
        output_file.write(text)


class IntentClassGenerator:
    """ Tools to automatically generate intent classes from an assistant
        definition.
    """

    def __init__(self):
        """ Initialisation. """
        self.jinja_env = Environment(
            loader=PackageLoader('snipsskills', 'templates'))

    def generate_intent_file(self, intent, output_dir):
        """ Given a JSON intent, generate the corresponding Python intent class
            file.

        :param intent: a JSON intent.
        """
        template = self.jinja_env.get_template('intent_template.py')
        # pylint: disable=no-member
        file_content = template.render(intent=intent)
        filename = camel_case_to_underscore(
            to_camelcase_capitalized(intent["name"])) + "_intent.py"
        save_intent_file(output_dir, filename, file_content)

    def generate_intent_registry_file(self, intents, output_dir):
        """ Given a list of intents, generate an intents registry, which is
            a list of intent classes.

        :param intents: a list of intents.
        """
        template = self.jinja_env.get_template('intent_registry_template.py')
        # pylint: disable=no-member
        file_content = template.render(intents=intents)
        with open("{}/intent_registry.py".format(output_dir), "w") as output_file:
            output_file.write(file_content)

    def generate(self, assistant_filename, output_dir):
        """ Generate intent classes from assistant.json specification.

        :param assistant_filename: path to the assistant zip file
        :param output_dir: directory to which the intents and registry should be
                           written.
        """
        self.jinja_env.globals.update(
            camel_case_to_underscore=camel_case_to_underscore)
        self.jinja_env.globals.update(
            to_camelcase_capitalized=to_camelcase_capitalized)

        intents = []
        content = zipfile.ZipFile(assistant_filename).read(
            'assistant/assistant.json')
        data = json.loads(content.decode('utf-8'))
        for intent in data["intents"]:
            self.generate_intent_file(intent, output_dir)
            intents.append(intent)

        self.generate_intent_registry_file(intents, output_dir)

        with open("{}/__init__.py".format(output_dir), "w") as output_file:
            output_file.write("")

        with open("{}/intents/__init__.py".format(output_dir), "w") as output_file:
            output_file.write("")



# -*-: coding utf-8 -*-
""" Auto-generated intent class. """

# *****************************************************************************
# *****************************************************************************
# *****************************************************************************
#
# WARNING: THIS IS AN AUTO-GENERATED FILE
# DO NOT ATTEMPT TO EDIT IT, AS CHANGES WILL BE OVERWRITTEN.
#
# *****************************************************************************
# *****************************************************************************
# *****************************************************************************

# pylint: disable=line-too-long

from snipsskillscore.intent_parser import IntentParser

class {{to_camelcase_capitalized(intent.name)}}Intent:

    intentName = "{{ intent.name }}"

    {% if intent.slots is defined and intent.slots|length > 0 -%}
    def __init__(self{% for slot in intent.slots -%}, {{slot.name}}=None{% endfor %}):
        {% for slot in intent.slots -%}
        self.{{slot.name}} = {{slot.name}}
        {% endfor %}
    {% endif -%}

    @staticmethod
    def parse(payload):
        intentName = IntentParser.get_intent_name(payload)
        if intentName != {{to_camelcase_capitalized(intent.name)}}Intent.intentName:
            return None
        return {{to_camelcase_capitalized(intent.name)}}Intent(
            {% for slot in intent.slots -%}
            IntentParser.get_slot_value(payload, "{{ slot.name}}"){{"," if not loop.last}}
            {% endfor -%}
        )



# -*-: coding utf-8 -*-
""" Utilities for handline Snipsfiles. """

import pkgutil
import os
import yaml

from ..models.skilldef import SkillDef
from ..models.intentdef import IntentDef


class SnipsfileParseException(Exception):
    """ Snipsfile parse exception class. """
    pass


class SnipsfileNotFoundError(Exception):
    """ Snipsfile not found error class. """
    pass


class SnipsspecNotFoundError(Exception):
    """ Snipsspec not found error class. """
    pass

# pylint: disable=too-few-public-methods


def get(yaml_config, key_path, default_value=None):
    """ Get a value in a yaml_config, or return a default value.

    :param yaml_config: the YAML config.
    :param key: a key to look for. This can also be a list, looking at
                more depth.
    :param default_value: a default value to return in case the key is
                          not found.
    :return: the value at the given key path, or default_value if not found.
    """
    if len(key_path) == 0:
        return default_value

    node = yaml_config
    for key in key_path:
        try:
            node = node[key]
        except Exception:
            return default_value
    return node


def find_intent(intent_name, intent_defs):
    """ Find an intent by name in a list of intent definitions.

    :param intent_name: the name of the intent to look for.
    :param intent_defs: a list of intent definitions.
    :return: the intent_def with matching name, or None.
    """
    for intent_def in intent_defs:
        if intent_def.name == intent_name:
            return intent_def
    return None

# pylint: disable=too-many-instance-attributes,too-many-locals


class Snipsfile:
    """ Utilities for handling Snipsfiles. """

    def __init__(self, snipsfile="Snipsfile"):
        """ Initialisation.

        :param snipsfile: name of the Snipsfile.
        """
        self.assistant_url = None

        if not os.path.isfile(snipsfile):
            raise SnipsfileNotFoundError(
                'No Snipsfile found at path {}. Please create one.'.format(snipsfile))

        yaml_config = None
        with open(snipsfile, 'r') as yaml_file:
            try:
                yaml_config = yaml.load(yaml_file)
            except yaml.scanner.ScannerError as err:
                raise SnipsfileParseException("Error parsing Snipsfile: " +
                                              str(err))

        if not yaml_config:
            return

        self.assistant_url = get(yaml_config, ['assistant'])
        if not self.assistant_url:
            raise SnipsfileParseException("No assistant definitions found.")

        self.snips_sdk_version = get(yaml_config, ['snips_sdk', 'version'])
        self.locale = get(yaml_config, ['locale'], 'en_US')
        self.tts_service = get(yaml_config, ['tts', 'service'])
        self.logging = get(yaml_config, ['logging'], True)
        self.default_location = get(
            yaml_config, ['default_location'], 'Paris,fr')
        self.mqtt_hostname = get(
            yaml_config, ['mqtt_broker', 'hostname'], 'localhost')
        self.mqtt_port = get(yaml_config, ['mqtt_broker', 'port'], 9898)
        
        self.microphone_config = MicrophoneConfig(yaml_config)

        self.skilldefs = []
        for skill in get(yaml_config, ['skills'], []):
            package_name = get(skill, ['package_name'])
            pip = get(skill, ['pip'])
            requires_tts = get(skill, ['requires_tts'], False)
            params = {}
            for key, value in get(skill, ['params'], {}).items():
                params[key] = value

            try:
                snipsspec_file = SnipsSpec(package_name)
            except (SnipsspecNotFoundError, SnipsfileParseException):
                snipsspec_file = None

            class_name = self.get_class_name(skill, snipsspec_file)
            intent_defs = self.get_intent_defs(skill, snipsspec_file)

            self.skilldefs.append(SkillDef(package_name, class_name, pip,
                                           params, intent_defs, requires_tts))

    def get_class_name(self, skill, snipsspec_file):
        """ Get the class name of a skill. The value, if provided, by the Snipsfile
            takes precedence over that of the Snipsspec file.

        :param skill: the skill def, as extracted from the Snipsfile.
        :param snipsspec_file: a SnipsSpec object, holding a fallback value
                               for the class name.
        :return: the class name of the skill.
        """
        class_name = get(skill, ['class_name'])
        if class_name is not None:
            return class_name
        if snipsspec_file is not None:
            try:
                return snipsspec_file.class_name
            except AttributeError:
                pass
        return None

    def get_intent_defs(self, skill, snipsspec_file):
        """ Get the intent definitions for a skill. The definitions for the
            skills found in skill has precendence over those in the
            snipsspec_file definitions, which act as fallbacks.


        :param skill: the skill def, as extracted from the Snipsfile.
        :param snipsspec_file: a SnipsSpec object, holding a fallback list of
                               intents.
        :return: the list of intents for the skill.
        """
        intents_snipsfile = []
        for intent in get(skill, ['intents'], []):
            name = get(intent, ['intent'])
            action = get(intent, ['action'])
            intents_snipsfile.append(IntentDef(name, action))

        if snipsspec_file is None:
            return intents_snipsfile

        try:
            intents_snipsspec = snipsspec_file.intent_defs
        except AttributeError:
            return intents_snipsfile

        intents = []
        for intent in intents_snipsfile:
            intents.append(intent)

        for intent in intents_snipsspec:
            found = find_intent(intent.name, intents_snipsfile)
            if not found:
                intents.append(intent)
        return intents


# pylint: disable=too-many-instance-attributes,too-many-locals
class SnipsSpec:
    """ Utilities for handling Snipsfiles. """

    def __init__(self, package_name):
        """ Initialisation.

        :param package_name: the name of the skills package, in which to look
                             for a Snipsspec file.
        """
        try:
            data = pkgutil.get_data(package_name, 'Snipsspec')
        except IOError:
            raise SnipsspecNotFoundError('No Snipsspec found.')

        if data is None:
            raise SnipsspecNotFoundError('No Snipsspec found.')

        yaml_config = None
        try:
            yaml_config = yaml.load(data)
        except yaml.scanner.ScannerError as err:
            raise SnipsfileParseException("Error parsing Snipsfile: " +
                                          str(err))

        if not yaml_config:
            return

        self.class_name = get(yaml_config, ['class_name'])

        self.intent_defs = []
        for intent in get(yaml_config, ['intents'], []):
            name = get(intent, ['intent'])
            action = get(intent, ['action'])
            self.intent_defs.append(IntentDef(name, action))


# pylint: disable=too-many-instance-attributes,too-many-locals
class MicrophoneConfig:
    """ Config holder for microphone. """

    def __init__(self, yaml_config):
        """ Initialisation.

        :param yaml_config: the YAML configuration
        """
        self.identifier = get(yaml_config, ['microphone','identifier'])
        self.params = {}
        for key, value in get(yaml_config, ['microphone','params'], {}).items():
            self.params[key] = value



# -*-: coding utf-8 -*-
""" snipsskills

Usage:
  snipsskills install
  snipsskills install [--snipsfile=<path>]
  snipsskills install bluetooth
  snipsskills run
  snipsskills run [--snipsfile=<path>]
  snipsskills -h | --help
  snipsskills --version

Options:
  -h --help                         Show this screen.
  --version                         Show version.
  --snipsfile=<path>                Path to the Snipsfile.

Examples:
  snipsskills install

Help:
  For help using this tool, please open an issue on the Github repository:
  https://github.com/snipsco/snipsskills
"""


from docopt import docopt

from . import __version__ as VERSION

def main():
    """ Main entry point. """
    options = docopt(__doc__, version=VERSION)

    if options['install'] == True:
        from snipsskills.commands.install import Install
        Install(options).run()
        return
    elif options['run'] == True:
        from snipsskills.commands.run import Run
        Run(options).run()
        return



# -*-: coding utf-8 -*-
""" Downloader for Snips assistants. """

import os

try:
    from urllib.request import urlopen
except ImportError:
    from urllib2 import urlopen

class AssistantDownloaderException(Exception):
    """ Exceptions related to downloads of Snips assistants. """
    pass

# pylint: disable=too-few-public-methods
class AssistantDownloader:
    """ Downloader for Snips assistants. """

    @staticmethod
    def download(url, output_dir, assistant_filename):
        """ Download an assistant, and save it to a file.

        :param url: the assistant URL.
        :param output_dir: the directory where the assistant should be
                           saved.
        """
        try:
            response = urlopen(url)
        except Exception:
            raise AssistantDownloaderException()

        AssistantDownloader.save_assistant(response.read(),
                                           output_dir, assistant_filename)

    @staticmethod
    def save_assistant(content, output_dir, assistant_filename):
        """ Save content of an assisant.

        :param content: the content of the file to save.
        :param output_dir: the directory where the assistant should be
                           saved.
        """
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
        output_filename = "{}/{}".format(output_dir, assistant_filename)
        with open(output_filename, "wb") as output_file:
            output_file.write(content)



# -*-: coding utf-8 -*-
""" Bluetooth setup utilities. """

import getpass
import os
import subprocess
import time

from snipsskillscore.logging import log, log_warning

from .os_helpers import cmd_exists, download_file, execute_command, remove_file, ask_yes_no, which
from .systemd import Systemd

SNIPSBLE_SERVICE_NAME = "snipsble"

# DO NOT CHANGE
SNIPS_BLE_SERVICE_UUID = "13EA4259-9D9E-42D1-A78B-638ED22CC768"
SNIPS_BLE_CHARACTERISTIC_UUID = "81D97A06-7A2D-4A98-A2E2-41688E3D8283"


class Bluetooth:
    """ Bluetooth setup utilities. """

    @staticmethod
    def setup(mqtt_hostname, mqtt_port):
        """ Setting up Bluetooth Relay MQTT service. """
        if ask_yes_no("Would you like to enable Bluetooth for this device?") == False:
            return

        try:
            Bluetooth.install_node()
        except Exception:
            log_warning("Could not download Node, which is required for Bluetooth. " +
                        "Please install Node manually, and restart the snipsskills installation script.")
            return

        Bluetooth.install_mqtt_relay()
        Bluetooth.setup_systemd(mqtt_hostname, mqtt_port)

    @staticmethod
    def install_node():
        """ Install node using dpkg, if it is not installed. """
        if cmd_exists('node'):
            return

        log("Node is required for Bluetooth setup. Installing Node.")

        filename = "node_latest_armhf.deb"

        log("Downloading Node.")
        download_file(
            "http://node-arm.herokuapp.com/node_latest_armhf.deb", filename)

        log("Installing Node.")
        execute_command("sudo dpkg -i node_latest_armhf.deb")
        remove_file(filename)

    @staticmethod
    def install_mqtt_relay():
        """ Install snips-mqtt-relay. """
        if not cmd_exists('node'):
            return

        log("Installing Node module: snips-mqtt-relay.")
        execute_command(
            "npm install --no-cache --prefix=./.snips snips-mqtt-relay &> /dev/null")

    @staticmethod
    def setup_systemd(mqtt_hostname, mqtt_port):
        (snipsble_path, node_path) = Bluetooth.get_params()
        contents = Systemd.get_template(SNIPSBLE_SERVICE_NAME)
        if contents is None:
            return
        contents = contents.replace("{{SNIPSBLE_PATH}}", snipsble_path) \
            .replace("{{SNIPS_BLE_SERVICE_UUID}}", SNIPS_BLE_SERVICE_UUID) \
            .replace("{{SNIPS_BLE_CHARACTERISTIC_UUID}}", SNIPS_BLE_CHARACTERISTIC_UUID) \
            .replace("{{SNIPS_MQTT_HOSTNAME}}", mqtt_hostname) \
            .replace("{{SNIPS_MQTT_PORT}}", str(mqtt_port)) \
            .replace("{{NODE_PATH}}", str(node_path))
        Systemd.write_systemd_file(SNIPSBLE_SERVICE_NAME, None, contents)
        Systemd.enable_service(None, SNIPSBLE_SERVICE_NAME)

    @staticmethod
    def get_params():
        snipsble_path = "{}/.snips/node_modules/snips-mqtt-relay".format(
            os.getcwd())
        node_path = which('node')
        return (snipsble_path, node_path)



# -*-: coding utf-8 -*-
"""The run command."""
# pylint: disable=too-few-public-methods,import-error

import os
import subprocess
import time
import threading

from sys import path

from ..utils.snipsfile_parser import Snipsfile, SnipsfileParseException, \
    SnipsfileNotFoundError
from ..utils.snips import Snips, SnipsNotFound, SnipsRuntimeFailure

from snipsskillscore.logging import log, log_success, log_warning, log_error
from snipsskillscore.server import Server
from snipsskillscore.instant_time import InstantTime
from snipsskillscore.time_interval import TimeInterval

from .base import Base, SNIPSFILE

path.append(".snips/intents")
path.append(".snips/intents/intents")

# pylint: disable=wrong-import-position,wrong-import-order
from intent_registry import IntentRegistry
# pylint: disable=wildcard-import,wrong-import-position,wrong-import-order
from intents import *

BINDINGS_FILE = "bindings.py"
INTENT_REGISTRY_FILE = ".snips/intents/intent_registry.py"

import threading


class Run(Base):
    """The run command."""

    # pylint: disable=undefined-variable,exec-used,eval-used
    def run(self):
        """ Command runner. """

        snipsfile_path = self.options['--snipsfile']
        if snipsfile_path is None or len(snipsfile_path) == 0:
            snipsfile_path = SNIPSFILE

        try:
            self.snipsfile = Snipsfile(snipsfile_path)
        except SnipsfileNotFoundError:
            log_error("Snipsfile not found. Please create one.")
            return
        except SnipsfileParseException as err:
            log_error(err)
            return

        registry = IntentRegistry()
        server = Server(self.snipsfile.mqtt_hostname,
                        self.snipsfile.mqtt_port,
                        self.snipsfile.logging,
                        self.snipsfile.tts_service,
                        self.snipsfile.locale,
                        registry,
                        self.handle_intent)

        self.skills = {}
        for skilldef in self.snipsfile.skilldefs:
            module_name = skilldef.package_name + "." + skilldef.package_name
            exec("from {} import {}".format(module_name, skilldef.class_name))
            cls = eval(skilldef.class_name)
            try:
                if skilldef.requires_tts:
                    skill_instance = cls(
                        tts_service=server.tts_service, **skilldef.params)
                else:
                    skill_instance = cls(**skilldef.params)
                self.skills[skilldef.package_name] = skill_instance
            except Exception as e:
                log_warning("Error loading skill {}: {}".format(
                    skilldef.package_name, str(e)))

        log("Starting the Snips Skills server.")
        server.start()

    def handle_intent(self, intent):
        thread = threading.Thread(target=self.handle_intent_async,
                                  args=(intent,))
        thread.start()

    def handle_intent_async(self, intent):
        """ Handle an intent.

        :param intent: the incoming intent to handle.
        """
        for skilldef in self.snipsfile.skilldefs:
            intent_def = skilldef.find(intent)
            if intent_def is None:
                continue
            skill = self.skills[skilldef.package_name]
            if intent_def.action.startswith("{%"):
                # Replace variables in scope with random variables
                # to prevent the skill from accessing/editing them.
                action = intent_def.action \
                    .replace("{%", "") \
                    .replace("%}", "") \
                    .replace("skilldef", "_snips_eejycfyrdfzilgfb") \
                    .replace("intent_def", "_snips_jkqdruouzuahmgns") \
                    .replace("snipsfile", "_snips_pdzdcpaygyjklngz") \
                    .replace("tts_service", "_snips_bxzbomfguxlyxswo") \
                    .strip()
                exec(action)
            else:
                getattr(skill, intent_def.action)()
